(**************************************************************************)
(*  Librairie pour un sujet de Khôlle de MP2I                             *)
(*  Triangulation de delaunay, Arbres bicolores sur les arêtes            *)
(*  Copyright (C) 2022                                                    *)
(*  Samy Jaziri                                                           *)
(*                                                                        *)
(*  La triangulation de delaunay est prise de la librairie ocamlgraph     *)
(*  La librairie semble ne pas marcher pour des ensembles trop gros       *)
(*  (10000 points) ?                                                      *)
(**************************************************************************)

(**************************************************************************)
(*                                                                        *)
(*  Ocamlgraph: a generic graph library for OCaml                         *)
(*  Copyright (C) 2004-2010                                               *)
(*  Sylvain Conchon, Jean-Christophe Filliatre and Julien Signoles        *)
(*                                                                        *)
(*  This software is free software; you can redistribute it and/or        *)
(*  modify it under the terms of the GNU Library General Public           *)
(*  License version 2.1, with the special exception on linking            *)
(*  described in file LICENSE.                                            *)
(*                                                                        *)
(*  This software is distributed in the hope that it will be useful,      *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *)
(*                                                                        *)
(**************************************************************************)

(* $Id: delaunay.ml,v 1.12 2005-11-02 13:43:35 filliatr Exp $ *)

(** Code follows Don Knuth's algorithm
    from ``Axioms and hulls'' (LNCS 606, Springer-Verlag, 1992), pp. 73-77.

    Some code and comments are taken from the Stanford Graph Base,
    file [gb_plane].
*)

module type CCC = sig
  type point
  val ccw : point -> point -> point -> bool
  val in_circle : point -> point -> point -> point -> bool
end

module type Triangulation = sig
  module S : CCC
  type triangulation
  val triangulate : S.point array -> triangulation
  val iter : (S.point -> S.point -> unit) -> triangulation -> unit
  val fold : (S.point -> S.point -> 'a -> 'a) -> triangulation -> 'a -> 'a
  val iter_triangles : 
    (S.point -> S.point -> S.point -> unit) -> triangulation -> unit
end

module Make (S : CCC) = struct

  module S = S

  type point = Point of int | Infinity

  (* Each edge of the current triangulation is represented by two arcs
     pointing in opposite directions; the two arcs are called mates. Each
     arc conceptually has a triangle on its left and a mate on its right. *)

  type arc = { 
    mutable vert : point; 
      (* v, if this arc goes from u to v *)
    mutable next : arc; 
      (* the arc from v that shares a triangle with this one *)
    mutable inst : node ref;
      (* instruction to change when the triangle is modified *)
    mate : int 
  }
  and node = 
    | Branch of int * int * node ref * node ref
    | Terminal of arc

  type triangulation = { 
    points : S.point array; 
    arcs : arc array;
    last_used_arc : int 
  }

  let rec dummy_arc = 
    { vert = Infinity; next = dummy_arc; 
      inst = ref (Terminal dummy_arc); mate = -1 }

  let make_arc n i =
    { vert = Infinity; next = dummy_arc; 
      inst = ref (Terminal dummy_arc); mate = 6 * n - 7 - i }

  let finite = function Point p -> p | Infinity -> assert false

  (* [flip] will be used in both steps T4 and T5 *)
  let flip c d e t'' p n n' =
    let e' = e.next in
    let c' = c.next in
    let c'' = c'.next in
    e.next <- c;
    c.next <- c'';
    c''.next <- e;
    c''.inst <- n; c.inst <- n; e.inst <- n;
    c.vert <- Point p;
    d.next <- e';
    e'.next <- c'; 
    c'.next <- d;
    c'.inst <- n'; e'.inst <- n'; d.inst <- n';
    d.vert <- Point t''

  let triangulate points =
    let ccw p q r = S.ccw points.(p) points.(q) points.(r) in
    let in_circle p q r s = 
      S.in_circle points.(p) points.(q) points.(r) points.(s)
    in
    let n = Array.length points in
    if n < 2 then invalid_arg "triangulate";
    let arcs = Array.init (6 * n - 6) (make_arc n) in
    let mate i = 6 * n - 7 - i in

    (* initialization: 
       create a trivial triangulation for the first 2 vertices *)
    let u = 0 in
    let v = 1 in
    let a1 = arcs.(0) in
    let a2 = arcs.(1) in
    let a3 = arcs.(2) in
    let b1 = arcs.(mate 0) in
    let b2 = arcs.(mate 1) in
    let b3 = arcs.(mate 2) in
    let l1 = ref (Terminal a2) in
    let l2 = ref (Terminal b3) in
    a1.vert <- Point v;  a1.next <- a2; a1.inst <- l1;
    a2.vert <- Infinity; a2.next <- a3; a2.inst <- l1;
    a3.vert <- Point u;  a3.next <- a1; a3.inst <- l1;
    b1.vert <- Point u;  b1.next <- b3; b1.inst <- l2;
    b2.vert <- Point v;  b2.next <- b1; b2.inst <- l2;
    b3.vert <- Infinity; b3.next <- b2; b3.inst <- l2;
    let l0 = ref (Branch (u, v, l1, l2)) in
    let j = ref 2 in (* last used arc *)

    (* then for each new vertex [p] *)
    for p = 2 to n - 1 do
      (* Step T1 *)
      let rec step_T1 l p = match !l with
	| Terminal al -> 
	    l, al
	| Branch (pl, ql, al, bl) -> 
	    step_T1 (if ccw pl ql p then al else bl) p
      in
      let l, al = step_T1 l0 p in

      (* Step T2 *)
      let a = al in
      let b = a.next in
      let c = b.next in
      let q = a.vert in
      let r = b.vert in
      let s = c.vert in
      j := !j + 3;
      let aj = arcs.(!j) in
      let aj_1 = arcs.(!j - 1) in
      let aj_2 = arcs.(!j - 2) in
      let bj = arcs.(aj.mate) in
      let bj_1 = arcs.(aj_1.mate) in
      let bj_2 = arcs.(aj_2.mate) in
      let l' = ref (Terminal a) in
      let l'' = ref (Terminal aj) in
      let l''' = ref (Terminal c) in
      aj.vert   <- q;         aj.next <- b;      aj.inst <- l'';
      aj_1.vert <- r;       aj_1.next <- c;    aj_1.inst <- l''';
      aj_2.vert <- s;       aj_2.next <- a;    aj_2.inst <- l';
      bj.vert   <- Point p;   bj.next <- aj_2;   bj.inst <- l';
      bj_1.vert <- Point p; bj_1.next <- aj;   bj_1.inst <- l'';
      bj_2.vert <- Point p; bj_2.next <- aj_1; bj_2.inst <- l''';
      a.next <- bj;   a.inst <- l'; 
      b.next <- bj_1; b.inst <- l'';
      c.next <- bj_2; c.inst <- l''';
      let r = finite r in
      let s = finite s in

      (* steps T3 or T4 depending on [q] *)
      let r = match q with
	| Point q -> (* Step T3 *)
	    let n = ref (Branch (q, p, l', l'')) in
	    let n' = ref (Branch (s, p, l''', l')) in
	    l := Branch (r, p, n, n');
	    r
	| Infinity -> (* Step T4 *)
	    let n = ref (Branch (s, p, l''', l')) in
	    l := Branch (r, p, l'', n);
	    let rec loop m a d s t =
	      if t <> r && ccw p s t then begin
		let n = ref (Terminal d) in
		match !m with
		  | Branch (mu, mv, ml, is_l') -> 
		      assert (is_l' == l');
		      m := Branch (mu, mv, ml, d.inst);
		      d.inst := Branch (t, p, n, l');
		      let m = d.inst in
		      flip a arcs.(a.mate) d t p n l';
		      let a = arcs.(a.mate).next in
		      let d = arcs.(a.mate).next in
		      let s = t in
		      let t = finite d.vert in
		      l' := Terminal a;
		      loop m a d s t
		  | Terminal _ -> 
		      assert false
	      end else begin
		(* at exit of while loop *)
		let n = ref (Terminal d.next) in
		d.inst := Branch (s, p, n, l');
		d.inst <- n; d.next.inst <- n; d.next.next.inst <- n;
		s
	      end
	    in
	    let d = arcs.(a.mate).next in
	    loop n a d s (finite d.vert)
      in

      (* Step T5 *)
      let rec loop c =
	let d = arcs.(c.mate) in
	let e = d.next in
	let t = finite d.vert in
	let t' = finite c.vert in
	let t'' = e.vert in
	if t'' <> Infinity && in_circle (finite t'') t' t p then begin
	  let t'' = finite t'' in
	  let n = ref (Terminal e) in
	  let n' = ref (Terminal d) in
	  c.inst := Branch (t'', p, n, n');
	  d.inst := Branch (t'', p, n, n');
	  flip c d e t'' p n n';
	  loop e
	end else if t' <> r then
	  loop arcs.(c.next.mate).next
	else
	  () (* break *)
      in
      loop c

    done;
    { points = points; arcs = arcs; last_used_arc = !j }

  let iter f t =
    let points = t.points in
    let n = Array.length t.arcs in
    for i = 0 to t.last_used_arc do
      match t.arcs.(i).vert, t.arcs.(n - 1 - i).vert with
	| Point u, Point v -> f points.(u) points.(v)
	| _ -> ()
    done

  let iter_triangles f t =
    let n = Array.length t.arcs in
    let seen_arc = Array.make n false in
    let mate i = n - 1 - i in
    let index a = mate a.mate in
    for i = 0 to n-1 do
      if not seen_arc.(i) then begin
	let a1 = t.arcs.(i) in
	let a2 = a1.next in
	let a3 = a2.next in
	seen_arc.(i) <- true;
	seen_arc.(index a2) <- true;
	seen_arc.(index a3) <- true;
	match a1.vert, a2.vert, a3.vert with
	  | Point i1, Point i2, Point i3 -> 
	      f t.points.(i1) t.points.(i2) t.points.(i3)
	  | _ ->
	      ()
      end
    done

  let fold f t a =
    let points = t.points in
    let n = Array.length t.arcs in
    let rec loop i a =
      if i <= t.last_used_arc then
	match t.arcs.(i).vert, t.arcs.(n - 1 - i).vert with
	  | Point u, Point v -> loop (succ i) (f points.(u) points.(v) a)
	  | _ -> loop (succ i) a
      else
	a
    in
    loop 0 a

end

(** Points with floating point coordinates *)

module FloatPoints = struct

  type point = float * float

  let ( + ) = ( +. )
  let ( - ) = ( -. )
  let ( * ) = ( *. )

  let det = function
    | [| [| a00; a01 |];
	 [| a10; a11 |] |] -> 
	a00 * a11 - a01 * a10
    | [| [| a00; a01; a02 |];
	 [| a10; a11; a12 |];
	 [| a20; a21; a22 |] |] -> 
	a00*a11*a22 - a00*a12*a21 - a10*a01*a22 + 
	a10*a02*a21 + a20*a01*a12 - a20*a02*a11
    | [| [| a00; a01; a02; a03 |];
	 [| a10; a11; a12; a13 |];
	 [| a20; a21; a22; a23 |];
	 [| a30; a31; a32; a33 |] |] -> 
	a00*a11*a22*a33 - a00*a11*a23*a32 - a00*a21*a12*a33 + 
	a00*a21*a13*a32 + a00*a31*a12*a23 - a00*a31*a13*a22 - 
	a10*a01*a22*a33 + a10*a01*a23*a32 + a10*a21*a02*a33 - 
	a10*a21*a03*a32 - a10*a31*a02*a23 + a10*a31*a03*a22 + 
	a20*a01*a12*a33 - a20*a01*a13*a32 - a20*a11*a02*a33 + 
	a20*a11*a03*a32 + a20*a31*a02*a13 - a20*a31*a03*a12 - 
	a30*a01*a12*a23 + a30*a01*a13*a22 + a30*a11*a02*a23 - 
	a30*a11*a03*a22 - a30*a21*a02*a13 + a30*a21*a03*a12
    | _ -> assert false

  let ccw (xu,yu) (xv,yv) (xw,yw) = 
    det [| [| xu; yu; 1.0 |];
	   [| xv; yv; 1.0 |];
	   [| xw; yw; 1.0 |] |] > 0.0

  let in_circle (xt,yt) (xu,yu) (xv,yv) (xw,yw) =
    det [| [| xt; yt; (xt * xt + yt * yt); 1.0 |];
	   [| xu; yu; (xu * xu + yu * yu); 1.0 |];
	   [| xv; yv; (xv * xv + yv * yv); 1.0 |];
	   [| xw; yw; (xw * xw + yw * yw); 1.0 |]; |] > 0.0

end

module Float = Make(FloatPoints)

(** Points with integer coordinates.
    We approximate using module [FloatPoints] but this could be made exact
    following Knuth's code in Axioms and Hulls *)

module IntPoints = struct

  type point = int * int

  let ccw (xu,yu) (xv,yv) (xw,yw) = 
    FloatPoints.ccw 
      (float xu, float yu) (float xv, float yv) (float xw, float yw)

  let in_circle (xt,yt) (xu,yu) (xv,yv) (xw,yw) = 
    FloatPoints.in_circle
      (float xt, float yt)
      (float xu, float yu) (float xv, float yv) (float xw, float yw)

end

module Int = Make(IntPoints)


(**************************************************************************)
(*  Arbres bicolores                                                                      *)
(*  Copyright (C) 2022                                                    *)
(*  Samy Jaziri                                                           *)
(***************************************************************************)

type point = IntPoints.point ;;
type 'a ensemble = 'a array ;;
type segment = point * point ;;
type arete = { segment : segment ;
               mutable zone_droite : int ;
               mutable zone_gauche : int } ;;
type sommet = { point : point ;
                mutable sortantes : arete list ;
                mutable entrantes : arete list }
type ordre = PLUS_PETIT | PLUS_GRAND | EGAUX ;;
type 'a abr = Vide | Noeud of 'a * 'a abr * 'a abr ;;
type couleur = ROUGE | NOIR ;;
type 'a arn = ( 'a * couleur ) abr ;;

(* Ordre des entiers classique *)
let ordre_int ( n : int ) ( m :int ) : ordre =
  if n < m then PLUS_PETIT
  else if n > m then PLUS_GRAND
  else EGAUX ;;

(* Ordre lexicographique surl les points *)
let ordre_lex ( point_g : point ) ( point_d : point ) : ordre =
  let xg, yg = point_g in
  let xd, yd = point_d in
  let ordre_x = ordre_int xg xd in
  match ordre_x with
  | EGAUX -> ordre_int yg yd
  | _ -> ordre_x ;;

(* Un pre-ordre pour les segments verticaux. Total dans le cas des segments *)
(* coupant une dalle *)
let rec ordre_lex_segments ( sg1 : segment ) ( sg2 : segment ) : ordre =
  let ordre_inverse ( o : ordre ) : ordre = match o with
    | PLUS_PETIT -> PLUS_GRAND
    | PLUS_GRAND -> PLUS_PETIT
    | EGAUX -> EGAUX in
  if sg1 = sg2 then EGAUX else begin
    let ( p1, _ ), ( p2, _ ) = ( sg1, sg2 ) in
    let ordre_x = ordre_lex p1 p2 in
    match ordre_x with
    | PLUS_GRAND -> ordre_inverse ( ordre_lex_segments sg2 sg1 )
    | _ -> begin
      let ( p1, q1 ) = sg1 in
      let ( p2, q2 ) = sg2 in
      let ordre_y = if IntPoints.ccw p1 q1 p2 then PLUS_PETIT
        else ( if IntPoints.ccw q1 p1 p2 then PLUS_GRAND else EGAUX ) in
      let ordre_v = if IntPoints.ccw q2 p2 q1 then PLUS_PETIT
        else ( if IntPoints.ccw p2 q2 q1 then PLUS_GRAND else EGAUX ) in
      match ordre_y with
      | EGAUX -> ordre_v
      | _  -> ordre_y
    end
  end ;;

(* Extension aux aretes *)
let ordre_lex_aretes ( a1 : arete ) ( a2 : arete ) : ordre =
  let ordre_s = ordre_lex_segments a1.segment a2.segment in
  match ordre_s with
  | EGAUX -> if a1.zone_droite = a2.zone_gauche then PLUS_GRAND else
    if a1.zone_gauche = a2.zone_droite then PLUS_PETIT else EGAUX
  | _ -> ordre_s ;;

(* Extension aux sommets *)
let ordre_lex_sommets ( s1 : sommet ) ( s2 : sommet ) : ordre =
  ordre_lex s1.point s2.point ;;

(* Comparaison entre un point et une arete *)
let ordre_point_arete ( p : point ) ( a : arete ) : ordre =
  let p1, p2 = a.segment in
  if ( IntPoints.ccw p1 p2 p ) then PLUS_GRAND
  else if ( IntPoints.ccw p p2 p1 ) then PLUS_PETIT
  else EGAUX ;;

(* Triangulation de delaunay d'un e nsemble de point *)
let delaunay_triangulation ( e : point array ) : sommet ensemble =
  let triangulation = Int.triangulate e in
  let nb_sommets = ref 0 in
  let numero_triangle = ref 1 in
  let sommet_abr = ref Vide in
  let rec insertion_sortantes ( a : arete )  ( abr : sommet abr ) : sommet abr =
    match abr with
    | Vide -> let p1, p2 = a.segment in
        let s = { point = p1 ; sortantes = [a] ; entrantes = []  } in (
        incr nb_sommets; Noeud ( s, Vide, Vide ) )
    | Noeud ( s', abr_g, abr_d ) ->
        let p, _ = a.segment in
        match ordre_lex p s'.point with
        | PLUS_PETIT -> Noeud ( s', insertion_sortantes a abr_g, abr_d )
        | PLUS_GRAND -> Noeud ( s', abr_g, insertion_sortantes a abr_d )
        | EGAUX ->
            let rec insert_arete
                ( a : arete ) ( liste_arete : arete list ) : arete list =
                match liste_arete with
                | [] -> [a]
                | a' :: l ->
                    let ( _, p ), ( _, p' ) = ( a.segment, a'.segment ) in
                    match ordre_lex p p' with
                    | EGAUX -> (
                        a'.zone_droite <-
                          max a'.zone_droite a.zone_droite ;
                        a'.zone_gauche <-
                          max a'.zone_gauche a.zone_gauche ;
                        liste_arete )
                    | _ -> a' :: ( insert_arete a l ) in (
              s'.sortantes <- ( insert_arete a s'.sortantes ) ;
              abr ) in
  let rec insertion_entrantes ( a : arete )  ( abr : sommet abr ) : sommet abr =
    match abr with
    | Vide -> let p1, p2 = a.segment in
        let s = { point = p2 ; sortantes = [] ; entrantes = [a]  } in (
        incr nb_sommets; Noeud ( s, Vide, Vide ) )
    | Noeud ( s', abr_g, abr_d ) ->
        let _, s = a.segment in
        match ordre_lex s s'.point with
        | PLUS_PETIT -> Noeud ( s', insertion_entrantes a abr_g, abr_d )
        | PLUS_GRAND -> Noeud ( s', abr_g, insertion_entrantes a abr_d )
        | EGAUX ->
            let rec insert_arete
                ( a : arete ) ( liste_arete : arete list ) : arete list =
                match liste_arete with
                | [] -> [a]
                | a' :: l ->
                    let ( p, _ ), ( p', _ ) = ( a.segment, a'.segment ) in
                    match ordre_lex p p' with
                    | EGAUX -> (
                        a'.zone_droite <-
                          max a'.zone_droite a.zone_droite ;
                        a'.zone_gauche <-
                          max a'.zone_gauche a.zone_gauche ;
                        liste_arete )
                    | _ -> a' :: ( insert_arete a l ) in (
              s'.entrantes <- ( insert_arete a s'.entrantes ) ;
              abr ) in
  let insertion ( a : arete ) ( abr : sommet abr ) : sommet abr =
    insertion_entrantes a ( insertion_sortantes a abr ) in
  let elements_trie ( abr : sommet abr ) ( taille : int ) : sommet ensemble =
    let elements = Array.make taille { point = (0,0) ;
                                       sortantes = [] ;
                                       entrantes = [] } in
    let c = ref 0 in
    let rec parcours ( abr : sommet abr ) : unit =
      match abr with
      | Vide -> ()
      | Noeud ( s, abr_g, abr_d ) -> (
          parcours abr_g ;
          elements.(!c) <- s ;
          incr c ;
          parcours abr_d ) in
    parcours abr ;
    elements in
  let rec tri_et_numerotation ( u : point ) ( v : point ) ( w : point ) : unit =
    if ( IntPoints.ccw u v w  )
    || ( not ( IntPoints.ccw u v w  ) && not ( IntPoints.ccw v u w  ) )
    then
      let arete1 = if ordre_lex u v = PLUS_GRAND
        then { segment = (v,u) ;
              zone_gauche = 0 ;
              zone_droite = !numero_triangle }
        else { segment = (u,v) ;
              zone_gauche = !numero_triangle ;
              zone_droite = 0 } in
      let arete2 = if ordre_lex v w = PLUS_GRAND
        then { segment = (w,v) ;
              zone_gauche = 0 ;
              zone_droite = !numero_triangle }
        else { segment = (v,w) ;
              zone_gauche = !numero_triangle ;
              zone_droite = 0 } in
      let arete3 = if ordre_lex w u = PLUS_GRAND
        then { segment = (u,w) ;
              zone_gauche = 0 ;
              zone_droite = !numero_triangle }
        else { segment = (w,u) ;
              zone_gauche = !numero_triangle ;
              zone_droite = 0 } in
      begin
      sommet_abr :=
        insertion arete1 ( insertion arete2 ( insertion arete3 !sommet_abr ) ) ;
      incr numero_triangle end
    else tri_et_numerotation v u w in
  Int.iter_triangles tri_et_numerotation triangulation ;
  elements_trie !sommet_abr !nb_sommets ;;

(* Rotation droite d'un ABR, renvoie un nouvel ABR *)
(* Seulement deux nouveaux noeuds sont créés *)
let rotation_droite ( abr : 'a abr ) : 'a abr = match abr with
  | Noeud ( y, Noeud ( x, abr_A, abr_B ) , abr_C ) ->
      Noeud ( x, abr_A, Noeud ( y, abr_B, abr_C ) )
  | _ -> abr ;;

(* Rotation gauche d'un ABR, renvoie un nouvel ABR *)
(* Seulement deux nouveaux noeuds sont créés *)
let rotation_gauche ( abr : 'a abr ) : 'a abr = match abr with
  | Noeud ( x, abr_A, Noeud ( y, abr_B, abr_C ) ) ->
      Noeud ( y, Noeud ( x, abr_A, abr_B ) , abr_C )
  | _ -> abr ;;

(* Renvoie la couleur de la racine d'un arbre bicolore *)
(* Si l'arbre est Vide, renvoie NOIR *)
let couleur_racine ( arn : 'a arn ) : couleur = match arn with
  | Vide -> NOIR
  | Noeud ( ( _, c ), _, _ ) -> c ;;

(** Insertion dans un arbre bicolore **)

(* Application d'une seule règle d'équilibrage sur un noeud interne d'un arbre *)
(* bicolore sur lequel a eu lieu une insertion. *)
(* Renvoie un nouvel arbre bicolore. Au maximum 4 nouveau noeuds son créés. *)
let equilibrage_insertion_arn ( arn : 'a arn ) : 'a arn =
  match arn with
  | Vide -> Vide
    (* Cas (a.1) *)
  | Noeud ( ( a, NOIR ),
            Noeud ( ( b, ROUGE ),
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ),
                    b_abr_d ),
            Noeud ( ( d, ROUGE ),
                    d_abr_g,
                    d_abr_d ) )
      -> Noeud ( ( a, ROUGE ),
            Noeud ( ( b, NOIR ),
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ),
                    b_abr_d ),
            Noeud ( ( d, NOIR ),
                    d_abr_g,
                    d_abr_d ) )
    (* Cas (a.1) symmétrique *)
  | Noeud ( ( a, NOIR ),
            Noeud ( ( d, ROUGE ),
                    d_abr_g,
                    d_abr_d ),
            Noeud ( ( b, ROUGE ),
                    b_abr_g,
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ) ) )
      -> Noeud ( ( a, ROUGE ),
            Noeud ( ( d, NOIR ),
                    d_abr_g,
                    d_abr_d ) ,
            Noeud ( ( b, NOIR ),
                    b_abr_g,
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ) ) )
    (* Cas (a.2) *)
  | Noeud ( ( a, NOIR ),
            Noeud ( ( b, ROUGE ),
                    b_abr_g,
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ) ),
            Noeud ( ( d, ROUGE ),
                    d_abr_g,
                    d_abr_d ) )
      -> Noeud ( ( a, ROUGE ),
            Noeud ( ( b, NOIR ),
                    b_abr_g,
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ) ),
            Noeud ( ( d, NOIR ),
                    d_abr_g,
                    d_abr_d ) )
    (* Cas (a.2) symmétrique *)
  | Noeud ( ( a, NOIR ),
            Noeud ( ( d, ROUGE ),
                    d_abr_g,
                    d_abr_d ),
            Noeud ( ( b, ROUGE ),
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d ) ,
                    b_abr_g ) )
      -> Noeud ( ( a, ROUGE ),
            Noeud ( ( d, NOIR ),
                    d_abr_g,
                    d_abr_d ),
            Noeud ( ( b, NOIR ),
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ),
                    b_abr_g ) )
    (* Cas (b) *)
  | Noeud ( ( a, NOIR ),
            Noeud ( ( b, ROUGE ),
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ),
                    b_abr_d ),
            a_abr_d ) when ( couleur_racine a_abr_d ) = NOIR
      -> rotation_droite (
          Noeud ( ( a, ROUGE ),
              Noeud ( ( b, NOIR ),
                      Noeud ( ( c, ROUGE ),
                              c_abr_g,
                              c_abr_d  ),
                      b_abr_d ),
              a_abr_d ) )
    (* Cas (b) symmétrique *)
  | Noeud ( ( a, NOIR ),
            a_abr_g,
            Noeud ( ( b, ROUGE ),
                    b_abr_g,
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ) ) ) when ( couleur_racine a_abr_g ) = NOIR
      -> rotation_gauche (
          Noeud ( ( a, ROUGE ),
              a_abr_g,
              Noeud ( ( b, NOIR ),
                      b_abr_g,
                      Noeud ( ( c, ROUGE ),
                              c_abr_g,
                              c_abr_d  ) ) ) ) 
    (* Cas (b.2) *)
  | Noeud ( ( a, NOIR ),
            Noeud ( ( b, ROUGE ),
                    b_abr_g,
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ) ),
            a_abr_d ) when ( couleur_racine a_abr_d ) = NOIR
      -> rotation_droite (
          Noeud ( ( a, ROUGE ),
              rotation_gauche (
                Noeud ( ( b, ROUGE ),
                  b_abr_g,
                  Noeud ( ( c, NOIR ),
                          c_abr_g,
                          c_abr_d  ) ) ),
              a_abr_d ) )
    (* Cas (b.2) symmétrique *)
  | Noeud ( ( a, NOIR ),
            a_abr_g,
            Noeud ( ( b, ROUGE ),
                    Noeud ( ( c, ROUGE ),
                            c_abr_g,
                            c_abr_d  ),
                    b_abr_g ) ) when ( couleur_racine a_abr_g ) = NOIR
      -> rotation_gauche (
          Noeud ( ( a, ROUGE ),
              a_abr_g,
              rotation_droite (
                Noeud ( ( b, ROUGE ),
                        Noeud ( ( c, NOIR ),
                                c_abr_g,
                                c_abr_d  ),
                        b_abr_g ) ) ) )
  | _ -> arn ;;

(* Insertion d'un element a dans un arbre bicolore arn, trié selon l'ordre *)
(* comp. Renvoi un nouvel arn. Le nombre de noeuds copiés est en O(log(n)) *)
let insertion_arn ( comp : 'a -> 'a -> ordre ) ( a : 'a ) ( arn : 'a arn ) : 'a arn =
  let rec insertion_equilibrage ( arn : 'a arn ) : 'a arn =
    match arn with
    | Vide -> Noeud ( ( a, ROUGE ) , Vide, Vide )
    | Noeud ( ( a', c ), arn_g, arn_d ) ->
        match comp a a' with
        | PLUS_PETIT ->
          equilibrage_insertion_arn
            ( Noeud ( ( a' , c ),
                      ( insertion_equilibrage arn_g ),
                      arn_d ) )
        | PLUS_GRAND ->
          equilibrage_insertion_arn
            ( Noeud ( ( a' , c ),
                      arn_g,
                      ( insertion_equilibrage arn_d ) ) )
        | EGAUX -> arn in
  (* Coloration de la racine en noir *)
  let nouvel_arn = insertion_equilibrage arn in
  match nouvel_arn with
  | Noeud ( ( a', ROUGE ), abr_g, abr_d ) -> Noeud ( ( a', NOIR ), abr_g, abr_d )
  | _ -> nouvel_arn ;;

(** Suppression dans un arbre bicolore *)

(* Application des règles d'équilibrage sur un noeud interne d'un arbre *)
(* bicolore sur lequel a eu lieu une suppression dans le sous-arbre droit. *)
(* Toutes les règles ne necessitant pas de remonter au père du noeud sont *)
(* appliquées *)
(* Renvoie un nouvel arbre bicolore. Au maximum 4 nouveau noeuds sonr créés *)
(* par appel récursif. Au maximum un seul appel récursif est effectué. *)
let rec equilibrage_suppression_gauche_arn ( arn : 'a arn ) : 'a arn * bool =
  match arn with
  (* Cas a. *)
  | Noeud ( ( p, NOIR ),
            arn_n,
            Noeud ( ( s, NOIR ), arn_c, arn_d ) )
    when ( couleur_racine arn_c ) = NOIR
           && ( couleur_racine arn_d ) = NOIR ->
    Noeud ( ( p, NOIR ),
            arn_n,
            Noeud ( ( s, ROUGE ), arn_c, arn_d ) ), false
  (* Cas b. *)
  | Noeud ( ( p, NOIR ),
            arn_n,
            Noeud ( ( s, ROUGE ), arn_c, arn_d ) ) ->
    let nouvel_arbre_p, _ = equilibrage_suppression_gauche_arn
        ( Noeud ( ( p, ROUGE ), arn_n, arn_c ) ) in
    ( Noeud ( ( s, NOIR ), nouvel_arbre_p, arn_d ) ), true
  (* Cas c *)
  | Noeud ( ( p, ROUGE ),
            arn_n,
            Noeud ( ( s, NOIR ), arn_c, arn_d ) )
    when ( couleur_racine arn_c ) = NOIR
           && ( couleur_racine arn_d ) = NOIR ->
    Noeud ( ( p, NOIR ),
            arn_n,
            Noeud ( ( s, ROUGE ), arn_c, arn_d ) ), true
  (* Cas d *)
  | Noeud ( ( p, col ),
            arn_n,
            Noeud ( ( s, NOIR ),
                    arn_c,
                    Noeud ( ( d, ROUGE ), d_arn_g, d_arn_d  ) ) ) ->
    ( rotation_gauche ( Noeud ( ( p, NOIR ),
                                arn_n,
                                Noeud ( ( s, col ),
                                        arn_c,
                                        Noeud ( ( d, NOIR ),
                                                d_arn_g,
                                                d_arn_d  ) ) ) ) ), true
  (* Cas e *)
  | Noeud ( ( p, _ ),
            arn_n,
            Noeud ( ( s, NOIR ),
                    Noeud ( ( c, ROUGE ), c_arn_g, c_arn_d  ),
                    arn_d ) ) ->
    ( equilibrage_suppression_gauche_arn
        ( Noeud ( ( p, NOIR ),
                  arn_n,
                  rotation_droite ( Noeud ( ( s, ROUGE ),
                                            Noeud ( ( c, NOIR ),
                                                    c_arn_g,
                                                    c_arn_d  ),
                                            arn_d ) ) ) ) )
  | _ -> arn, true ;;

(* Application des règles d'équilibrage sur un noeud interne d'un arbre *)
(* bicolore sur lequel a eu lieu une suppression dans le sous-arbre gauche. *)
(* Toutes les règles ne necessitant pas de remonter au père du noeud sont *)
(* appliquées *)
(* Renvoie un nouvel arbre bicolore. Au maximum 4 nouveau noeuds sonr créés *)
(* par appel récursif. Au maximum un seul appel récursif est effectué. *)
let rec equilibrage_suppression_droite_arn ( arn : 'a arn ) : 'a arn * bool =
  match arn with
  (* Cas a. *)
  | Noeud ( ( p, NOIR ),
            Noeud ( ( s, NOIR ), arn_c, arn_d ),
            arn_n )
    when ( couleur_racine arn_c ) = NOIR
           && ( couleur_racine arn_d ) = NOIR ->
    Noeud ( ( p, NOIR ),
            Noeud ( ( s, ROUGE ), arn_c, arn_d ),
            arn_n ), false
  (* Cas b. *)
  | Noeud ( ( p, NOIR ),
            Noeud ( ( s, ROUGE ), arn_c, arn_d ),
            arn_n ) ->
    let nouvel_arbre_p, _ = equilibrage_suppression_droite_arn
        ( Noeud ( ( p, ROUGE ), arn_d, arn_n ) ) in
    ( Noeud ( ( s, NOIR ), arn_c, nouvel_arbre_p ) ), true
  (* Cas c *)
  | Noeud ( ( p, ROUGE ),
            Noeud ( ( s, NOIR ), arn_c, arn_d ),
            arn_n )
    when ( couleur_racine arn_c ) = NOIR
           && ( couleur_racine arn_d ) = NOIR ->
    Noeud ( ( p, NOIR ),
            Noeud ( ( s, ROUGE ), arn_c, arn_d ),
            arn_n ), true
  (* Cas d *)
  | Noeud ( ( p, col ),
            Noeud ( ( s, NOIR ),
                    Noeud ( ( c, ROUGE ), c_arn_g, c_arn_d  ),
                    arn_d ),
            arn_n ) ->
    ( rotation_droite ( Noeud ( ( p, NOIR ),
                                Noeud ( ( s, col ),
                                        Noeud ( ( c, NOIR ),
                                                c_arn_g,
                                                c_arn_d  ),
                                        arn_d ),
                                arn_n ) ) ), true
  (* Cas e *)
  | Noeud ( ( p, _ ),
            Noeud ( ( s, NOIR ),
                    arn_c,
                    Noeud ( ( d, ROUGE ), d_arn_g, d_arn_d  ) ),
            arn_n ) ->
    ( equilibrage_suppression_droite_arn
        ( Noeud ( ( p, NOIR ),
                  rotation_gauche ( Noeud ( ( s, ROUGE ),
                                            arn_c,
                                            Noeud ( ( d, NOIR ),
                                                d_arn_g,
                                                d_arn_d  ) ) ),
                  arn_n ) ) )
  | _ -> arn, true ;;

(* Suppression d'un element a dans un arbre bicolore arn, trié selon l'ordre *)
(* comp. Renvoi un nouvel arn. Le nombre de noeuds copiés est en O(log(n)) *)
let suppression_arn ( comp : 'a -> 'a -> ordre ) ( a : 'a ) ( arn : 'a arn ) : 'a arn =
  let rec suppression_equilibrage ( arn : 'a arn ) : 'a arn * bool =
    match arn with
    | Vide -> Vide, true
    | Noeud ( ( a', c ), arn_g, arn_d ) ->
      match comp a a' with
      | PLUS_PETIT ->
        let sous_arn, equilibre = suppression_equilibrage arn_g in
        let nouvel_arbre = Noeud ( ( a' , c ), sous_arn, arn_d ) in
        if equilibre then nouvel_arbre, true
        else equilibrage_suppression_gauche_arn nouvel_arbre
      | PLUS_GRAND ->
        let sous_arn, equilibre = suppression_equilibrage arn_d in
        let nouvel_arbre = Noeud ( ( a' , c ), arn_g, sous_arn ) in
        if equilibre then nouvel_arbre, true
        else equilibrage_suppression_droite_arn nouvel_arbre
      | EGAUX ->
        match arn_g with
        | Vide ->
          begin
            match arn_d with
            | Noeud ( ( a'', ROUGE ), arn_d_g, arn_d_d ) ->
              Noeud ( ( a'', NOIR ), arn_d_g, arn_d_d ), true
            | _ -> arn_d, c = ROUGE
          end
        | _ -> let rec predecesseur ( sous_arn : 'a arn ) : 'a arn * bool * 'a =
                match sous_arn with
                | Vide -> failwith "Appel de predecesseur sur arbre gauche vide"
                | Noeud ( ( a'', c' ), arn_d_g, Vide ) ->
                  ( arn_d_g, c' = ROUGE, a'' )
                | Noeud ( ( a'', c' ), arn_g_g, arn_g_d ) ->
                  let sous_arn, equilibre, remplacant = predecesseur arn_g_d in
                  let nouvel_arbre =
                    Noeud ( ( a'' , c' ), arn_g_g, sous_arn ) in
                  if equilibre then nouvel_arbre, true, remplacant
                  else let nouvel_arbre_equilibre, equilibre =
                         ( equilibrage_suppression_droite_arn nouvel_arbre ) in
                    nouvel_arbre_equilibre, equilibre, remplacant in
          let sous_arn, equilibre, remplacant = predecesseur arn_g in
          let nouvel_arbre = Noeud ( ( remplacant, c ), sous_arn, arn_d ) in
          if equilibre then nouvel_arbre, true
          else equilibrage_suppression_gauche_arn nouvel_arbre in
  (* Coloration de la racine en noir *)
  let nouvel_arn, _ = suppression_equilibrage arn in
  nouvel_arn ;;
